#include <iostream>
#include <string>
#include <fstream>
#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <cassert>
#include <time.h>

#include "../Lib/liblbfgs-1.8/include/lbfgs.h"
#include "ToolBayesK41Reg.h"
#include "CImg.h"
#include "../Lib/Corpetti/io_field.h"

#include      "../Lib/Spline/coeff.h"
#include      "../Lib/Spline/interpol.h"
#include      "../Lib/Spline/derivee_spline.h"
#include      "../Lib/Spline/interpol.h"
#include      "../Lib/Spline/io.h"

#define ABS(n)   ((n > 0) ? (n) : -(n))

using namespace cimg_library;
using namespace std;

/*************************************************************************************************************/

/*************************************************************************************************************/
/**********************************Spline Interpolation  and Derivative Filters ******************************/
/*************************************************************************************************************/

/*************************************************************************************************************/
void Recal_spline(CImg<double> &out,const CImg<double> &in,const CImg<double> &UV, int SplineDegree){
    
    cout<<"\n Recalage"<<endl;
    short int nb_col=in.dimx();
    short int nb_lig=in.dimy();
    float *Im_in;
    float *Im_out=new float[nb_col*nb_lig];
    float *u;
    float *v;
	
    Im_in=CImg2float(in);
    u=CImg2float(UV.get_slice(0));
    v=CImg2float(UV.get_slice(1));
    
    float	*ImageRasterArray=(float*)calloc(nb_col*nb_lig,sizeof(float));
    Copy_f(ImageRasterArray,Im_in,nb_col,nb_lig);
    if (SplineDegree>2){bool Error = SamplesToCoefficients(ImageRasterArray, nb_col, nb_lig, SplineDegree); // on recupere le coefficient
        if(Error){printf("probleme\n");}
    }
    
    short i,j;
    int I_u,I_v;
    float Du,Dv;
    float U,V;
    
    float *ad;
    float p00,p01,p10,p11;
    int D;
    short inc_i,inc_j;
    float pix;float * ptImout;float * ptImin;float * ptU;float * ptV;
    ptImout=Im_out;ptImin=Im_in;ptU=u;ptV=v;
    
    for(i=0;i<nb_lig;i++)
        for(j=0;j<nb_col;j++,ptImin++,ptImout++,ptU++,ptV++){
            U=(*ptU);
            V=(*ptV);
            if(U>0)
                inc_j=1;
            else
                inc_j=-1;
            
            if(V>0)
                inc_i=1;
            else
                inc_i=-1;
            
            I_u=(int)(U);
            I_v=(int)(V);
            
            Du=ABS(U)-ABS(I_u);
            Dv=ABS(V)-ABS(I_v);
            
            if((float)j+U<0.0)
            { I_u=-j;
                Du=0.0;
            }
            
            if((float)j+U >(float)(nb_col-1))
            { I_u=nb_col-1-j;
                Du=0.0;
            }
            
            if( (float)i+V<0.0)
            {	I_v=-i;
                Dv=0.0;
            }
            
            if((float)i+V>(float)(nb_lig-1))
            {	I_v=nb_lig-1-i;
                Dv=0.0;
            }
            
            D=I_u + (I_v * nb_col);
            ad=ptImin + D;
            
            p11=Du*Dv;
            p01=Du*(1-Dv);
            p10=(1-Du)*Dv;
            p00=(1-Du)*(1-Dv);
            
            /* -- bords images --*/
            if((I_u+j+inc_j)>=nb_col || (I_u+j+inc_j)<0 || (I_v+i+inc_i)>=nb_lig || (I_v+i+inc_i)<0){
                pix = (*ad);
            }
            /* -- interieur images --*/
            else{
                if(abs(in(I_u+j+inc_j,I_v+i+inc_i))>=1022||abs(in(I_u+j,I_v+i))>=1022||abs(in(I_u+j+inc_j,I_v+i))>=1022||abs(in(I_u+j,I_v+i+inc_i))>=1022
                   /*||(in(I_u+j+inc_j,I_v+i+inc_i))==0||(in(I_u+j,I_v+i))==0||(in(I_u+j+inc_j,I_v+i))==0||(in(I_u+j,I_v+i+inc_i))==0*/){
                    pix = (*ad);
                }
                else{
                    if (SplineDegree>2){
                        pix = (float)InterpolatedValue(ImageRasterArray, nb_col, nb_lig, j+U, i+V, SplineDegree);
                    }
                    else{
                        pix = (*ad)*p00 +
                        (*(ad+inc_j))*p01 +
                        (*(ad+inc_i*nb_col))*p10 +
                        (*(ad+inc_j+inc_i*nb_col))*p11;
                    }
                }
            }
            *ptImout= pix;
        }
    out=float2CImg(out.dimx(),out.dimy(), Im_out);
    free(ImageRasterArray);
    delete [] Im_in;delete [] Im_out;delete [] u;delete [] v;
}
void Proj_Dup_Spline(CImg<double> &out, const CImg<double> &in, int SplineDegree){
    
	short int nb_col=in.dimx();
	short int nb_lig=in.dimy();
	float *I_low;
	float *I_high=new float[nb_col*nb_lig*4];
	I_low=CImg2float(in);
	float	*ImageRasterArray=(float*)calloc(nb_col*nb_lig,sizeof(float));
	Copy_f(ImageRasterArray,I_low,nb_col,nb_lig);
	bool Error = SamplesToCoefficients(ImageRasterArray, nb_col, nb_lig, SplineDegree); // on recupere le coefficient
	if(Error){ printf("probleme\n"); }
	
    float *pt;
   	short L_s;
   	short i,j;
	int count=0;
    
    L_s=nb_col<<1;
    pt=I_high;
    
    for(i=0;i<(nb_lig-1);i++,pt+=L_s,count+=L_s)
    {
 	   	for(j=0;j<(nb_col-1);j++,pt+=2,count+=2)
 	   	{
            *(pt) = (float)InterpolatedValue(ImageRasterArray, nb_col, nb_lig, j, i, SplineDegree);
            *(pt+L_s) = (float)InterpolatedValue(ImageRasterArray, nb_col, nb_lig, j, i+0.5, SplineDegree);
            *(pt+1) = (float)InterpolatedValue(ImageRasterArray, nb_col, nb_lig, j+0.5, i, SplineDegree);
            *(pt+1+L_s) = (float)InterpolatedValue(ImageRasterArray, nb_col, nb_lig, j+0.5, i+0.5, SplineDegree);
 	   	}
		//bord droit
	   	*(pt) = (float)InterpolatedValue(ImageRasterArray, nb_col, nb_lig, j, i, SplineDegree);
		*(pt+L_s) =(float)InterpolatedValue(ImageRasterArray, nb_col, nb_lig, j, i+0.5 , SplineDegree);
		*(pt+1) = *(pt);
	   	*(pt+1+L_s) =  *(pt+L_s);
		pt+=2;count+=2;
    }
    //derniere ligne
    for(j=0;j<(nb_col-1);j++,pt+=2,count+=2)
    {
        *(pt) = (float)InterpolatedValue(ImageRasterArray, nb_col, nb_lig, j, i, SplineDegree);
        *(pt+L_s) = *(pt);
        *(pt+1) = (float)InterpolatedValue(ImageRasterArray, nb_col, nb_lig, j+0.5, i, SplineDegree);
        *(pt+1+L_s) = *(pt+1);
    }
    //coin droite-bas
    *(pt) = (float)InterpolatedValue(ImageRasterArray, nb_col, nb_lig, j, i, SplineDegree);
    *(pt+L_s) = *(pt);
    *(pt+1) = *(pt);
    *(pt+1+L_s) = *(pt);
    out=float2CImg(out.dimx(),out.dimy(), I_high);
    free(ImageRasterArray);
    delete [] I_high;delete [] I_low;
    
}
CImg<double> gradSpline_X(const CImg<double> &in, float sigma){
    
    int SplineDegree=3;
    short int nb_col=in.dimx();
    short int nb_lig=in.dimy();
    CImg<double> out(nb_col,nb_lig);
    CImg<double> tmp_1(nb_col,nb_lig);
    CImg<double> tmp0(nb_col,nb_lig);
    CImg<double> tmp1(nb_col,nb_lig);
    float *u;
    float *coef=(float*)calloc(nb_col*nb_lig, sizeof(float));
    if (sigma!=0){out=in.get_blur(sigma,true);}else{out=in;}
    u=CImg2float(out);
    if (directBsplineMirror(u,coef, nb_col, nb_lig, 1,SplineDegree ) != 0){exit(1);}
    gradX(coef,u,nb_col,nb_lig, SplineDegree,0);
    out=float2CImg(out.dimx(),out.dimy(), u);
    free(coef);
    delete [] u;
    return out;
    
}
CImg<double> gradSpline_Y(const CImg<double> &in, float sigma){
    
    int SplineDegree=3;
    short int nb_col=in.dimx();
    short int nb_lig=in.dimy();
    CImg<double> out(nb_col,nb_lig);
    CImg<double> tmp_1(nb_col,nb_lig);
    CImg<double> tmp0(nb_col,nb_lig);
    CImg<double> tmp1(nb_col,nb_lig);
    float *u;
    float *coef=(float*)calloc(nb_col*nb_lig, sizeof(float));
    if (sigma!=0){out=in.get_blur(sigma,true);}else{out=in;}
    u=CImg2float(out);
    if (directBsplineMirror(u,coef, nb_col, nb_lig, 1,SplineDegree ) != 0){exit(1);}
    gradY(coef,u,nb_col,nb_lig, SplineDegree,0);
    out=float2CImg(out.dimx(),out.dimy(), u);
    free(coef);
    delete [] u;
    return out;
    
}
CImg<double> gradSpline_XX(const CImg<double> &in, float sigma){
    
    int SplineDegree=3;
    short int nb_col=in.dimx();
    short int nb_lig=in.dimy();
    CImg<double> out(nb_col,nb_lig);
    float *u;
    if (sigma!=0){out=in.get_blur(sigma,true);}else{out=in;}
    u=CImg2float(out);
    float *coef=(float*)calloc(nb_col*nb_lig, sizeof(float));
    if ( directBsplineMirror(u,coef, nb_col, nb_lig, 1,SplineDegree ) != 0){exit(1);}
    derivXX(coef,u,nb_col,nb_lig, SplineDegree,0);
    out=float2CImg(out.dimx(),out.dimy(), u);
    free(coef);
    delete [] u;
    return out;
    
}
CImg<double> gradSpline_YY(const CImg<double> &in, float sigma){
    
    int SplineDegree=3;
    short int nb_col=in.dimx();
    short int nb_lig=in.dimy();
    CImg<double> out(nb_col,nb_lig);
    float *u;
    if (sigma!=0){out=in.get_blur(sigma,true);}else{out=in;}
    u=CImg2float(out);
    float *coef=(float*)calloc(nb_col*nb_lig, sizeof(float));
    if ( directBsplineMirror(u,coef, nb_col, nb_lig, 1,SplineDegree ) != 0){exit(1);}
    derivYY(coef,u,nb_col,nb_lig, SplineDegree,0);
    out=float2CImg(out.dimx(),out.dimy(), u);
    free(coef);
    delete [] u;
    return out;
    
}
CImg<double> gradSpline_XY(const CImg<double> &in, float sigma){
    
    int SplineDegree=3;
    short int nb_col=in.dimx();
    short int nb_lig=in.dimy();
    CImg<double> out(nb_col,nb_lig);
    float *u;
    if (sigma!=0){out=in.get_blur(sigma,true);}else{out=in;}
    u=CImg2float(out);
    float *coef=(float*)calloc(nb_col*nb_lig, sizeof(float));
    if ( directBsplineMirror(u,coef, nb_col, nb_lig, 1,SplineDegree ) != 0){exit(1);}
    derivXY(coef,u,nb_col,nb_lig, SplineDegree,0);
    out=float2CImg(out.dimx(),out.dimy(), u);
    free(coef);
    delete [] u;
    return out;
    
}
CImg<double> grad_t(const CImg<double> &in,const CImg<double> &indt, float sigma){
	
    short int nb_col=in.dimx();
    short int nb_lig=in.dimy();
    CImg<double> out(nb_col,nb_lig);
    
    if (sigma!=0){
        out=(indt-in).get_blur(sigma,true);
    }
    else{out=indt-in;}
    return out;
}

/*************************************************************************************************************/

/*************************************************************************************************************/
/********************************** Ecriture champs (procedure Thomas)   *************************************/
/*************************************************************************************************************/

/*************************************************************************************************************/

void write_aMotionFile(char *fileMotionOut, const CImg<double>& im){
	
	int dx=im.dimx();
   	int dy=im.dimy();
  	float *ima_u=new float[dx*dy];
	float *ima_v=new float[dx*dy];
	for (int j=0;j<dy;j++){
        for (int i=0;i<dx;i++){
			ima_u[j*dx+i]=im(i,j,0);
			ima_v[j*dx+i]=im(i,j,1);
        }
	}
	cout<<fileMotionOut<<endl;
	write_champ(ima_u, ima_v, dx, dy,fileMotionOut);
	delete [] ima_u;
	delete [] ima_v;
    
}
void write_XAmotionH(char *fileXafOut, const CImg<double>& im){
	
	int dx=im.dimx();
   	int dy=im.dimy();
	short *error=new short[1];
  	float *ima_u=new float[dx*dy];
	float *ima_v=new float[dx*dy];
	for (int j=0;j<dy;j++){
        for (int i=0;i<dx;i++){
			ima_u[j*dx+i]=(float)im(i,j,0);
			ima_v[j*dx+i]=(float)im(i,j,1);
        }
	}
	cout<<"writing "<<fileXafOut<<endl;
	write_field(fileXafOut,ima_u, ima_v, dx, dy, error);
	
	delete [] ima_u;
	delete [] ima_v;
}
CImg<float> read_aMotionFile(char * filename,int dx,int dy,int Dx, int ofx, int ofy){
    
	int ii,jj;
	float u,v;
	FILE * file=NULL;
	int readInt;
	char * readLine =new char[200];
    
	CImg<float> im(dx,dy,2);
	CImg<float> aux(Dx,dy+ofy, 2);
    
	
	assert (file=fopen(filename,"r"));
	cout<<filename<<endl;
	fscanf (file,"%d %d\n",&readInt,&readInt);
	fgets(readLine,100,file);
	
	for (int j=0;j<dy+ofy;j++){
		for (int i=0;i<Dx;i++){
			fscanf(file,"%d %d %f %f\n",&ii,&jj,&u,&v);
			aux(i,j,0)=u;
			aux(i,j,1)=v;
		}
	}
    fclose(file);
	for (int j=0;j<dy;j++){
		for (int i=0;i<dx;i++){
			im(i,j,0)=aux(i+ofx,j+ofy,0);
			im(i,j,1)=aux(i+ofx,j+ofy,1);
		}
	}
    
	delete [] readLine;
    
	return 	im;
}

CImg<float> lire_XAmotionH(char * filename,int dx,int dy, int ofx, int ofy,int Dx,int Dy){
    
	
	short *error=new short[1];
	float *field_u=new float[Dx*Dy];
	float *field_v=new float[Dx*Dy];
    
	CImg<float> im(dx,dy,2);
    
	cout<<"reading file "<<filename<<endl;
	read_field(filename,field_u, field_v, (short) Dx,(short) Dy, error);
	for (int i=0;i<dx;i++){
		for (int j=0;j<dy;j++){
			im(i,j,0)=field_u[(j+ofy)*Dx+i+ofx];
			im(i,j,1)=field_v[(j+ofy)*Dx+i+ofx];
		}
	}
    
	
	delete [] field_u;
	delete [] field_v;
    
	return 	im;
}
CImg<float> read_aUVDavisFile(char * filename,int dx,int dy,int Dx, int ofx, int ofy, double Umoy){
    
	int ii,jj;
	float u,v;
	FILE * file=NULL;
	char * readLine =new char[200];
    
	CImg<float> im(dx,dy,2);
	CImg<float> aux(Dx,dy+ofy, 2);
	aux.fill(0.);
    
	
	assert (file=fopen(filename,"r"));
	cout<<filename<<endl;
	//fgets(readLine,120,file);
	fgets(readLine,120,file);
  	while(strcmp(readLine,"##\n")!=0){
        sscanf(readLine, "\t%d %d %f %f\n",&ii,&jj,&u,&v);
        if(jj<dy+ofy){
            aux(ii,jj,0)=u-Umoy;
            aux(ii,jj,1)=v;
        }
        fgets(readLine,120,file);
        
	}
    fclose(file);
	
	for (int i=0;i<dy;i++){
		for (int j=0;j<dx;j++){
			//im(i,dy-1-j,0)=-aux(i+ofx,j+ofy,0);
			//im(i,dy-1-j,1)=aux(i+ofx,j+ofy,1);
			im(i,j,0)=-aux(i+ofx,j+ofy,0);
			im(i,j,1)=-aux(i+ofx,j+ofy,1);
		}
	}
	
	delete [] readLine;
    
	return 	im;
}
void readLambdaInit(CImg <double> &lambda,CImg <double> &lambda_cross,char * filename){
    
	float lmda1,lmda2;
	FILE * file=NULL;
	assert (file=fopen(filename,"r"));
	cout<<filename<<endl;
	for (int j=0;j<lambda.dimx();j++){
        fscanf(file,"%f %f\n",&lmda1,&lmda2);
        lambda(j)=lmda1;
        lambda_cross(j)=lmda2;
		cout<<lambda(j)<<" "<<lambda_cross(j)<<endl;
	}
	fclose(file);
	
}


/*************************************************************************************************************/

/*************************************************************************************************************/
/**************************************************** Main   *************************************************/
/*************************************************************************************************************/

/*************************************************************************************************************/

int main(int argc, char *argv [])
{
    cimg_usage("Optic flow estimation with K41 regularization. \nWritten by Patrick Heas, 2012, Inria\nAll right reserved");
	
    cout<<"\n--------------------------------------------------------"<<endl;
    cout<<"---------------Input Parameters-------------------------"<<endl;
    cout<<"--------------------------------------------------------"<<endl;
    /*-- Parameters --*/
    string ifn = cimg_option("-inPath","inputBayesK41","input text file: \n\t\t\t\t\t\t 1-st line: path to first image file in tiff format \n\t\t\t\t\t\t 2-nd line: path to second image file in tiff format \n\t\t\t\t\t\t 3-rd line: path to output directory");
    int dy = cimg_option("-dy",256,"image line number");
    int dx  = cimg_option("-dx",256,"image column number");
    int neta  = cimg_option("-neta",0,"smallest scale for power-law constraint ");
    int lo  = cimg_option("-lo",4,"largest scale  for power-law constraint");
    double gamma  = cimg_option("-gamma",1.E-2,"factor power law");
    double power  = cimg_option("-power",1.,"exponent power law");
    
    
    cout<<"\n--------------------------------------------------------"<<endl;
    cout<<"---------------Options ---------------------------------"<<endl; 
    cout<<"--------------------------------------------------------"<<endl;
    cout<<"--Mutiresolution pyramid --------------------------------------------------------"<<endl;
    int nb_resol  = cimg_option("-multiresol",1,"number of resolution levels");
    int resolEnd= cimg_option("-resolEnd",1,"finest resolution level");
    cout<<"--Data model --------------------------------------------------------------------"<<endl;
    double norm= cimg_option("-norm",2.,"penalization for data term : 2=L2,1=L1,0=Leclerc");
    double sigma_obs= cimg_option("-sigma_obs",50.,"initial robust parameter control on data term (for norm L1 or leclerc)");
    double sigBlur = cimg_option("-sigBlur",0.,"amplitude Tchumperlé's isotropic blurring for image pre-smoothing");
    double sigDiff  = cimg_option("-sigDiff",0.,"passive scalar transport diffusion coefficient (instead of a simple OFC transport) ");
    int offEdge  = cimg_option("-offEdeges",0,"border marge suppression from the data term ");
    cout<<"--Optimization and Initialization------------------------------------------------ "<<endl;
    int nbIt = cimg_option("-nbItTrace",30,"nb iteration for trace randomization");
    double eps  = cimg_option("-eps",1.E-10,"precision for convergence of dual algorithm");
    int nbIt_lbfgs = cimg_option("-nbItLbfgs",10,"nb iteration max lbfgs");
    double Umoy  = cimg_option("-Umean",0.,"Initialization value for the average horizontal velocity");
    bool HS  = cimg_option("-HS",false,"Horn and Shunck multireoslution");
    bool initHS  = cimg_option("-initHS",false," HS motion field at coarser resolutions (1/2^j | j > 1)");
    double alpha  = cimg_option("-alpha",10.,"Horn & Schunck regularization coefficient (if used)");
    double lambdaInit  = cimg_option("-lambdaInit",1.,"Initial lagrangian multipliers value");
    double lambdaInitFile  = cimg_option("-lambdaInitFile",false,"Initial lagrangian multipliers values from file \n\t\t\t\t\t\t(file path must appear in the 5-th line of input text file given in argument of -inPath)");
    cout<<"--ROI (cropping)-----------------------------------------------------------------"<<endl;
    int ofy = cimg_option("-ofy",0,"offset line number ");
    int ofx  = cimg_option("-ofx",0,"offset column number");
    int Dx  = cimg_option("-Dx",-1,"whole Image column number");
    int Dy  = cimg_option("-Dy",-1,"whole Image line number");
    if (Dx<0){if(ofx==0)Dx=dx ; else {cout<<"Error in -ofx !! "<<endl;ofx=0;}}
    if (Dy<0){if(ofy==0)Dy=dy; else {cout<<"Error in -ofy !! "<<endl;ofy=0;}}
    cout<<"--MISC---------------------------------------------------------------------------- "<<endl;
    bool pgm  = cimg_option("-pgm",false,"image in pgm format (true) (instead of tiff format (false))");
    bool DivZero = cimg_option("-divZero",false,"divergence free projection at each resolution level (default false)");
    bool GT= cimg_option("-GT",false,"ground truth comparison (default=false) \n\t\t\t\t\t\t(true motion file path must appear in the 4-th line  of input text file given in argument of -inPath)");
    int offGT=cimg_option("-offGT",0," border dimx not considered for ground truth comparison");
    
    bool display=false;
    
    float scaleSmooth=0;
    double tol=1E-2;
    int nb_recal=1;
    int  netaFineResol  = neta;
    int  loFineResol  = lo;
    if (loFineResol<neta){loFineResol=neta;}
    int dxOrig=dx;
    int dyOrig=dy;
    int MaxAll;
    if (dxOrig<dyOrig){MaxAll=(int)(dxOrig/2.);}
    else{MaxAll=(int)(dyOrig/2.);}
    double gammaFineResol=gamma;
    double FactorRobParam=1;
    if (norm==2){FactorRobParam=1.;}
    if (norm==1){FactorRobParam=1./(2*sigma_obs);}
    if (norm==0){FactorRobParam=1./(sigma_obs*sigma_obs);}
    
    /*Reading images and creation of motion compensate image by mean field*/
    char **fileIm= alloc_seqPath(5);
    char ch[100];
    read_path(ifn,5,fileIm);
    CImg <double> imOrig(dxOrig,dyOrig);
    CImg <double> imdtOrig(dxOrig,dyOrig);
    CImg <double> imOrigLarge(Dx,2*Dy);
    CImg <double> tmpLarge(Dx,Dy);
    CImg <double> Umean(Dx,Dy);
    CImg <double> XOrig(dxOrig,dyOrig,2);
    Umean.fill(Umoy);
    if (pgm){
        imOrig=read_pgm(fileIm[0],dxOrig,dyOrig,ofx,ofy);
        if (Umoy!=0){Recal_spline(tmpLarge,read_pgm(fileIm[1],Dx,Dy,0,0),fill(Umean,Umean*0.),3);}
        else{tmpLarge=read_pgm(fileIm[1],Dx,Dy,0,0);}
        imdtOrig=tmpLarge.get_crop(ofx,ofy,ofx+dxOrig-1,ofy+dyOrig-1);
    }
    else{
        
        imOrig=imOrigLarge.get_load_tiff(fileIm[0]).get_slice(0).get_channel(0).get_crop(ofx,ofy,ofx+dxOrig-1,ofy+dyOrig-1);
        if (Umoy!=0){Recal_spline(tmpLarge ,imOrigLarge.get_load_tiff(fileIm[1]).get_slice(0).get_channel(0).get_crop(0,0,Dx-1,Dy-1),fill(Umean,Umean*0.),3);}
        else{tmpLarge=imOrigLarge.get_load_tiff(fileIm[1]).get_slice(0).get_channel(0).get_crop(0,0,Dx-1,Dy-1);}
        imdtOrig=tmpLarge.get_crop(ofx,ofy,ofx+dxOrig-1,ofy+dyOrig-1);
        
        
    }
    
    /*Expanding images for multiresolution*/
    expand_dim(imOrig.dimx(),imOrig.dimy(),nb_resol,&dx,&dy);
    cout<<"image size before "<<dxOrig<<" "<<dyOrig;
    cout<<", and after expanding "<<dx<<" "<<dy<<endl;
    CImg <double> im(dx,dy);
    CImg <double> imdt(dx,dy);
    im=expand(imOrig,nb_resol);
    imdt=expand(imdtOrig,nb_resol);
    sprintf(ch,"%sIm_t.pgm",fileIm[2]);
    write_pgm255(imOrig,ch,0,255);
    sprintf(ch,"%sIm_dt.pgm",fileIm[2]);
    write_pgm255(imdtOrig,ch,0,255);
    //if (display){im.display("im");imdt.display("imdt");}
    /*Memory allocation */
    CImg <double> im_x(dx,dy);
    CImg <double> im_y(dx,dy);
    CImg <double> im_t(dx,dy);
    CImg <double> data_out(dx,dy);
    CImgList<double> pyrIm=alloc_pyramid (dx,dy,1,nb_resol);
    CImgList<double> pyrImdt=alloc_pyramid (dx,dy,1,nb_resol);
    CImgList<double> pyrOut=alloc_pyramid (dx,dy,1,nb_resol);
    CImgList<double> pyrf_x=alloc_pyramid (dx,dy,1,nb_resol);
    CImgList<double> pyrf_y=alloc_pyramid (dx,dy,1,nb_resol);
    CImgList<double> pyrf_t=alloc_pyramid (dx,dy,1,nb_resol);
    CImgList<double> pyrf_norm=alloc_pyramid (dx,dy,1,nb_resol);
    
    CImg <double> X(dx,dy,2);
    CImg <double> trueUV(dx,dy,2);
    CImgList<double> pyrXtil=alloc_pyramid (dx,dy,2,nb_resol);
    CImgList<double> pyrXinc=alloc_pyramid (dx,dy,2,nb_resol);
    CImgList<double> pyrTmp=alloc_pyramid (dx,dy,1,nb_resol);
    CImgList<double> pyrDiv=alloc_pyramid (dx,dy,1,nb_resol);
    CImgList<double> pyrDiv_x=alloc_pyramid (dx,dy,1,nb_resol);
    CImgList<double> pyrDiv_y=alloc_pyramid (dx,dy,1,nb_resol);
    
    CImgList<double> pyrf_lambda=alloc_pyramid (loFineResol,nb_resol);
    CImgList<double> pyrf_lambda_cross=alloc_pyramid (loFineResol,nb_resol);
    
    double absConstDev;
    int nombre;
    double maxEvidenceBeta;
    double evidencePrior;
    double priorEnergy;
    
    CImg <double> save(20);
    save.fill(0.);
    
    
    
    /*ground truth*/
    trueUV.fill(0.);
    if (GT){
        //trueUV=read_aUVFile(fileIm[3],dx,dy,Dx,Dy,ofx,ofy);
        trueUV=read_aMotionFile(fileIm[3],dx,dy,Dx,ofx,ofy);
        sprintf(ch,"%s_true.motion",fileIm[2]);
        write_aMotionFile(ch,trueUV);
        sprintf(ch,"%s_trueS2_HV",fileIm[2]);
        write_S2(StructureFunction2nd_horiVertDir(trueUV,1,MaxAll-1,false),ch);
        sprintf(ch,"%s_trueS2_Cross",fileIm[2]);
        write_S2(StructureFunction2nd_crossDir(trueUV,1,MaxAll-1,false),ch);
    }
    /*-- Construction de la pyramid multiresolution  (image sparse ou non) --*/
    CImg<double> PB_burt(5);PB_burt(0,0)=0.05;PB_burt(1,0)=0.25; PB_burt(2,0)=0.4;PB_burt(3,0)=0.25; PB_burt(4,0)=0.05;
    pyrIm[0]=im;pyrImdt[0]=imdt;
    if (lambdaInitFile) {readLambdaInit(pyrf_lambda[0],pyrf_lambda_cross[0],fileIm[4]);}
    else{ pyrf_lambda[0].fill(lambdaInit); pyrf_lambda_cross[0].fill(lambdaInit);}
    for(int resol=0;resol<nb_resol-1;resol++){
        pyrIm[resol+1]=Filtre_2D_sparse(Filtre_2D_sparse(pyrIm[resol],PB_burt),PB_burt.get_transpose()).get_resize(pyrIm[resol+1],1,1);
        pyrImdt[resol+1]=Filtre_2D_sparse(Filtre_2D_sparse(pyrImdt[resol],PB_burt),PB_burt.get_transpose()).get_resize(pyrImdt[resol+1],1,1);
        pyrf_lambda[resol+1].fill(lambdaInit);pyrf_lambda_cross[resol+1].fill(lambdaInit);
        
    }
    pyrXtil[nb_resol-1].fill(0.);
    
    time_t tdeb, t1,t2;
    time(&tdeb);
    t1=tdeb;
    /* -- Multiresolution --*/
    bool alreadyRecal=false;
    int normSave=norm;
    for (int resol=nb_resol-1; resol>=resolEnd-1;resol--){
        gamma=gammaFineResol*pow((double)(pow(2.,(double)resol)),2.-power);
        cout<<"****************************************************************"<<endl;
        cout<<"Resolution "<<resol<<" (gamma: "<<gamma<<" zeta: "<<power<<")"<<endl;
        cout<<"****************************************************************"<<endl;
        neta=max((int)(netaFineResol/(pow(2.,(double)resol))),1);
        lo=max((int)(loFineResol/(pow(2.,(double)resol))),1);
        cout<<"Scale interval [neta,lo]=["<<neta<<","<<lo<<"]"<<endl;
        
        pyrf_x[resol]=gradSpline_X(pyrImdt[resol],sigBlur);
        pyrf_y[resol]=gradSpline_Y(pyrImdt[resol],sigBlur);
        pyrf_t[resol]=grad_t(pyrIm[resol],pyrImdt[resol],sigBlur);
        
        if (sigDiff!=0){
            pyrf_t[resol]-=0.5*gradSpline_XX(pyrIm[resol],sigBlur)*sigDiff*sigDiff;
            pyrf_t[resol]-=0.5*gradSpline_YY(pyrIm[resol],sigBlur)*sigDiff*sigDiff;
        }
        pyrf_norm[resol].fill(1.);
        pyrf_x[resol].div(pyrf_norm[resol]);
        pyrf_y[resol].div(pyrf_norm[resol]);
        pyrf_t[resol].div(pyrf_norm[resol]);
        if (!alreadyRecal){pyrOut[resol].fill(1.);}
        setZeroBorder(pyrOut[resol],offEdge,offEdge);if (resol==0){setZeroBorderOnExpandedRegion(pyrOut[resol],dxOrig,dyOrig);}
        pyrXinc[resol].fill(0.);
        
        if (HS || (initHS && resol >= 1 )){
            //Horn and Schunck on each resolution level
            HornAndSchunck(pyrXinc[resol],pyrXtil[resol],pyrf_x[resol],pyrf_y[resol],pyrf_t[resol],alpha,pyrOut[resol],false);
            for (int kk=0;kk<5;kk++){
                if (norm==1){pyrOut[resol]=RobustDataCoef_QuasiTV(pyrXinc[resol],pyrf_x[resol],pyrf_y[resol],pyrf_t[resol],sigma_obs);}
                if (norm==0){pyrOut[resol]=RobustDataCoef(pyrXinc[resol],pyrf_x[resol],pyrf_y[resol],pyrf_t[resol],sigma_obs);}
                setZeroBorder(pyrOut[resol],offEdge,offEdge); if (resol==0){setZeroBorderOnExpandedRegion(pyrOut[resol],dxOrig,dyOrig);}
                HornAndSchunck(pyrXinc[resol],pyrXtil[resol],pyrf_x[resol],pyrf_y[resol],pyrf_t[resol],alpha,pyrOut[resol],false);
            }
        }
        else{
            //autosimilar regularization on each resolution level
            cout<<"precision "<<eps*pow(2.,(double)resol)<<endl;
            if (initHS&&normSave!=2){norm=normSave;}
            //if (norm==1){pyrOut[resol]=RobustDataCoef_QuasiTV(pyrXinc[resol],pyrf_x[resol],pyrf_y[resol],pyrf_t[resol],sigma_obs);cout<<"*"<<endl;}
            //if (norm==0){pyrOut[resol]=RobustDataCoef(pyrXinc[resol],pyrf_x[resol],pyrf_y[resol],pyrf_t[resol],sigma_obs);cout<<"*"<<endl;}
            setZeroBorder(pyrOut[resol],offEdge,offEdge);if (resol==0){setZeroBorderOnExpandedRegion(pyrOut[resol],dxOrig,dyOrig);}
            nombre=NumberNonZeroElementInHessian_K41Long(pyrf_x[resol],pyrf_y[resol],pyrOut[resol],pyrf_lambda[resol],pyrf_lambda_cross[resol],scaleSmooth,neta,lo);
            cout<<"number non zero elements: "<<nombre<<endl;
            PrimalMinimizer_K41Long(pyrXinc[resol],pyrXtil[resol],pyrf_x[resol],pyrf_y[resol],pyrf_t[resol],pyrOut[resol],
                                    pyrf_lambda[resol],pyrf_lambda_cross[resol],scaleSmooth,neta,lo,nombre);
            //if(initHS){norm=2;}
            absConstDev=optimizeDualFunction_LongK41(pyrf_lambda[resol],pyrf_lambda_cross[resol],pyrXinc[resol],pyrXtil[resol],
                                                     trueUV.get_resize(pyrf_x[resol].dimx(),pyrf_x[resol].dimy(),2)/pow(2.,(double)resol),
                                                     pyrf_x[resol],pyrf_y[resol],pyrf_t[resol],pyrOut[resol],pyrOut[resol].get_fill(1.),
                                                     resol,nombre,offEdge,dxOrig,dyOrig,norm,sigma_obs,gamma,power,scaleSmooth,neta,lo,eps*pow(2.,(double)resol),nbIt_lbfgs,display,DivZero);
            save(10)=absConstDev;
            if (nbIt!=0){
                cout<<"\nBayesian evidence"<<endl;
                maxEvidenceBeta= evidence_betaLong(pyrXinc[resol],pyrf_x[resol],pyrf_y[resol],pyrf_t[resol],pyrOut[resol],
                                                   pyrf_lambda[resol],pyrf_lambda_cross[resol],scaleSmooth,neta,lo,nombre,nbIt,tol,FactorRobParam);cout<<".";fflush(stdout);
                evidencePrior=evidence_powerLawLong(pyrXinc[resol],pyrf_x[resol],pyrf_y[resol],pyrf_t[resol],pyrOut[resol],
                                                    maxEvidenceBeta, pyrf_lambda[resol],pyrf_lambda_cross[resol],scaleSmooth,neta,lo,nombre,FactorRobParam,save); cout<<".";fflush(stdout);
                priorEnergy=2*maxEvidenceBeta*DualFunction_K41Long(pyrXinc[resol],pyrXtil[resol],pyrf_x[resol],pyrf_y[resol],pyrf_t[resol],pyrOut[resol].get_fill(0.),
                                                                   pyrf_lambda[resol],pyrf_lambda_cross[resol],scaleSmooth,neta,lo,gamma,power,nombre);
                cout<<"max. evidence Beta = "<<maxEvidenceBeta<<endl;
                cout<<"Neglected prior energy: "<<priorEnergy<<endl;
            }
            if (GT){groundTruthComparison(((pyrXinc[resol]+pyrXtil[resol]).get_resize(trueUV.dimx(),trueUV.dimy(),2)*pow(2.,(double)resol)).get_crop(offGT,offGT,0,dxOrig-1-offGT,dyOrig-1-offGT,1),trueUV.get_crop(offGT,offGT,0,dxOrig-1-offGT,dyOrig-1-offGT,1),trueUV.get_crop(offGT,offGT,0,dxOrig-1-offGT,dyOrig-1-offGT,1).get_slice(0).get_fill(1.),fileIm[2],save);}
            if (GT){groundTruthComparison(pyrXinc[resol]+pyrXtil[resol],trueUV.get_resize(pyrXinc[resol].dimx(),pyrXinc[resol].dimy(),2)/pow(2.,(double)resol),pyrXinc[resol].get_slice(0).get_fill(1.),fileIm[2],save);}
            
        }
        
        if(resol!=0){
            Proj_Dup_Spline(pyrTmp[resol-1], (pyrXtil[resol].get_slice(0)+pyrXinc[resol].get_slice(0))*2.,3);
            FillChannel(pyrXtil[resol-1],pyrTmp[resol-1],0);
            Proj_Dup_Spline(pyrTmp[resol-1], (pyrXtil[resol].get_slice(1)+pyrXinc[resol].get_slice(1))*2.,3);
            FillChannel(pyrXtil[resol-1],pyrTmp[resol-1],1);
            if (DivZero){maj_vecteurs(pyrXtil[resol-1],pyrXtil[resol-1],true);}
            X=pyrXtil[resol-1].get_resize(X.dimx(),X.dimy(),2)*pow(2.,(double)(resol-1.));
            pyrTmp[resol-1]=pyrImdt[resol-1];
            Recal_spline(pyrImdt[resol-1],pyrTmp[resol-1],pyrXtil[resol-1],3);
        }
        else{
            if (nb_recal==1){
                if (DivZero){
                    sprintf(ch,"%sUVwithDiv_Gamma%eZeta%e.motion",fileIm[2],gamma,power);
                    write_aMotionFile(ch,(pyrXtil[0]+pyrXinc[0]).get_crop(0,0,dxOrig-1,dyOrig-1));
                    maj_vecteurs(X,pyrXtil[0]+pyrXinc[0],true);
                }
                else{X=pyrXtil[0]+pyrXinc[0];}
                sprintf(ch,"%sUV_Gamma%eZeta%e.motion",fileIm[2],gamma,power);
                write_aMotionFile(ch,X.get_crop(0,0,dxOrig-1,dyOrig-1));
                if (display){X.get_crop(0,0,0,dxOrig-1,dyOrig-1,1).display("estimate V");}
                sprintf(ch,"%sS2_HV_Gamma%eZeta%e",fileIm[2],gamma,power);
                write_S2(StructureFunction2nd_horiVertDir(X.get_crop(0,0,dxOrig-1,dyOrig-1),1,MaxAll-1,false),ch);
                sprintf(ch,"%sS2_Cross_Gamma%eZeta%e",fileIm[2],gamma,power);
                write_S2(StructureFunction2nd_crossDir(X.get_crop(0,0,dxOrig-1,dyOrig-1),1,MaxAll-1,false),ch);
                sprintf(ch,"%sDataOut_Gamma%eZeta%e.pgm",fileIm[2],gamma,power);
                write_pgm255(pyrOut[0],ch,0,1);	
            }
            else{
                alreadyRecal=true;
                pyrf_lambda[0].fill(lambdaInit); pyrf_lambda_cross[0].fill(lambdaInit);
                pyrXtil[0]=pyrXtil[0]+pyrXinc[0];
                if (DivZero){maj_vecteurs(pyrXtil[0],pyrXtil[0],true);}
                X= pyrXtil[0];pyrXinc[0].fill(0.);
                Recal_spline(pyrImdt[0],imdt,pyrXtil[0],3);
                nb_recal--;resol++;
                
            }	
        }
        if (GT){
            if (resol==0){groundTruthComparison(X.get_crop(offGT,offGT,0,dxOrig-1-offGT,dyOrig-1-offGT,1),trueUV.get_crop(offGT,offGT,0,dxOrig-1-offGT,dyOrig-1-offGT,1),trueUV.get_crop(offGT,offGT,0,dxOrig-1-offGT,dyOrig-1-offGT,1).get_slice(0).get_fill(1.),fileIm[2],save);}
        }
        save(0)=gamma;
        save(1)=power;
        save(8)=priorEnergy;
        save(9)=maxEvidenceBeta;
        sprintf(ch,"%sevidence_resol%d",fileIm[2],resol);
        writeLine(save,ch);
    }	 
    
    
    time(&t2);
    printf("Processing time duration (sec): %d\n",(int)(t2-t1));
    cout<<"\n ** End algorithm **"<<endl;	
    return 0;
}




















