#ifndef ToolBayesK41Reg_h
#define ToolBayesK41Reg_h

#include <iostream>
#include <string> 
#include <fstream>
#include <cstdio> 
#include <cstdlib>
#include <cmath>
#include <cassert>

#include "CImg.h"

#include "comprow_double.h"
#include "ilupre_double.h"
#include "icpre_double.h"

#include "diagpre_double.h"
#include "icpre_double.h"
#include "ilupre_double.h"
#include "cgs.h"

#include <../../Lib/liblbfgs-1.8/include/lbfgs.h>

using namespace cimg_library;
using namespace std;
 	/*************************************************************************************************************/

	/*************************************************************************************************************/
	/**********************************      Various functions   *************************************************/
	/*************************************************************************************************************/

	/*************************************************************************************************************/
void BlocDensify(CImg<double> &X, int sampInitx,int sampInity, int sampInt);
void writeLine(const CImg<double> &save,char * filename);
CImg<double> treshold_minmax(const CImg<double> &in, double treshmin, double treshmax, double flag);
CImg<float> read_aUVFile(char * filename,int dx,int dy,int Dx, int Dy,int ofx, int ofy);
char ** alloc_seqPath(int nbIm);
void read_path(string ifn1, int nbIm, char ** fileIm);
void maj_vecteurs(CImg<double> &X,const CImg<double> & Xin,bool divNul);
CImg<double> divImOrder2(const CImg<double> &im);
CImg<double> expand(const CImg<double> &imOrig,int nbResol);
void expand_dim(int nb_col,int nb_lig,int resol,int *nouv_nb_col,int *nouv_nb_lig);
CImg<short> read_pgm(char *nom,int dx,int dy,int ofx,int ofy);
void write_pgm255(const CImg<double> &in,char * filename, float min, float max);
CImgList<double> alloc_pyramid (int nc, int coarse);
CImgList<double> alloc_pyramid (int nc, int nl, int nslice, int coarse);
void FillChannel(CImg<double> &out , const  CImg<double> & im, int band);
float * CImg2float( const CImg<double> &in);
CImg<double> float2CImg(int dx, int dy, float *in);
void Copy_f(float *im1, float *im2, int nb_col, int nb_line);
CImg<double> fill(const CImg<double> &a,const CImg<double> &b);
void write_S2(const CImg<double> &S2, char * filename);
void write_SP(const CImg<double> &S1,const CImg<double> &S2,const CImg<double> &S3,const CImg<double> &S4,const CImg<double> &S5,const CImg<double> &S6,const CImg<double> &S7,const CImg<double> &S8,const CImg<double> &S9,const CImg<double> &S10, const CImg<double> &S11,const CImg<double> &S12,const CImg<double> &S13,const CImg<double> &S14,const CImg<double> &S15,const CImg<double> &S16,const CImg<double> &S17,const CImg<double> &S18,const CImg<double> &S19,const CImg<double> &S20,char * filename);
void groundTruthComparison(const CImg<double> &XOrig,const CImg<double> &trueUV,const CImg<double> &maskFlag,char * filename);
void groundTruthComparison(const CImg<double> &XOrig,const CImg<double> &trueUV,const CImg<double> &maskFlag,char * filename,CImg<double> &save);
	/*************************************************************************************************************/

	/*************************************************************************************************************/
	/**********************************      Border conditions  *************************************************/
	/*************************************************************************************************************/

	/*************************************************************************************************************/
void setZeroBorder(CImg<double> &a,int ofx, int ofy);
void setZeroBorderOnExpandedRegion(CImg<double> &a,int dxOrig, int dyOrig);
void fixPeriodConditions(int nc,int nl,CImg<double> &A1_i_1j,CImg<double> &A1_i1j,CImg<double> &A1_ij_1,CImg<double> &A1_ij1,
			 CImg<double> &A1_ijnc,CImg<double> &A1_ij0,CImg<double> &A1_inlj,CImg<double> &A1_i0j);
void fixNeumannConditions(int nc,int nl,CImg<double> &A1_i_1j,CImg<double> &A1_ij,CImg<double> &A1_i1j,CImg<double> &A1_ij_1,CImg<double> &A1_ij1);
void excludeWhite(int neta,int lo,CImgList<double> &A_iljl,CImgList<double> &A_i_lj_l,CImgList<double> &A_i_ljl,CImgList<double> &A_ilj_l);	
void fixDirichletConditions(int nc,int nl,CImgList<double> &A_i_lj,CImgList<double> &A_ij_l,CImgList<double> &A_ijl,CImgList<double> &A_ilj,int neta,int lo);	
void fixDirichletConditions(int nc,int nl,CImg<double> &A_i_1j,CImg<double> &A_ij,CImg<double> &A_i1j,CImg<double> &A_ij_1,CImg<double> &A_ij1);	
void fixDirichletConditions_cross(int nc,int nl,CImgList<double> &A_i_lj_l,CImgList<double> &A_i_ljl,CImgList<double> &A_ilj_l,CImgList<double> &A_iljl,int neta,int lo);
void fixDirichletConditions_cross(int nc,int nl,CImg<double> &A_ij,CImgList<double> &A_i_lj_l,CImgList<double> &A_i_ljl,CImgList<double> &A_ilj_l,CImgList<double> &A_iljl,int neta,int lo);
       /*************************************************************************************************************/

	/*************************************************************************************************************/
	/*****************************************Robust  Functions***************************************************/
	/*************************************************************************************************************/

	/*************************************************************************************************************/
void Data_Out_Gen_f_QuasiTV(const CImg<double>&u,const CImg<double>&v,const CImg<double>&fx,const CImg<double>&fy,const CImg<double>&ft,float s,int nb_col,int nb_line,CImg<double> &Data_out);
CImg <double> RobustDataCoef_QuasiTV(const CImg<double> &UV,const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> &im_t,double sigma);
void Data_Out_Gen_f(const CImg<double>&u,const CImg<double>&v,const CImg<double>&fx,const CImg<double>&fy,const CImg<double>&ft,float sigma,int nb_col,int nb_line,CImg<double> &Data_out);
CImg <double> RobustDataCoef(const CImg<double> &UV,const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> &im_t,double sigma);
void UpdateRobustWeight_Leclerc(const CImg<double> &UV, CImgList <double> &z, double tau,bool periodic);
void UpdateRobustWeight_QuasiTV(const CImg<double> &UV, CImgList <double> &z, double tau,bool periodic);

	/*************************************************************************************************************/

	/*************************************************************************************************************/
	/**********************************                  Filters *************************************************/
	/*************************************************************************************************************/

	/*************************************************************************************************************/
CImg<double>  Filtre_2D(const CImg<double>& b, const CImg<double>& filtre);
CImg<double>  Filtre_2D(const CImg<double>& b, const CImg<double>& filtre, bool periodic);
CImg<double>  Filtre_2D_sparse(const CImg<double>& b, const CImg<double>& filtre);
CImg<double>  Filtre_2D_sparse(const CImg<double>& b, const CImg<double>& filtre, bool periodic);
CImg<double>  Filtre_2D_sparse(const CImg<double>& b, const CImg<double>& filtre,float max);
CImg<double>  Filtre_2D_sparse(const CImg<double>& b, const CImg<double>& filtre, bool periodic,float max);
CImg<double> laplacian(const CImg<double> &im, bool periodic);
	/*************************************************************************************************************/

	/*************************************************************************************************************/
	/*****************************CGS Inversion, log det and trace -1*********************************************/
	/*************************************************************************************************************/

	/*************************************************************************************************************/
int standard_CGS_inversion(int taille, int nombre, double *val,int *row_ptr,int *col_ind, VECTOR_double B, CImg<double> &res,double tol);
int standard_CGS_inversion(int taille, int nombre, double *val,int *row_ptr,int *col_ind, VECTOR_double B, CImg<double> &res);	
double ILU_logDetA(int taille, int nombre, double *val,int *row_ptr,int *col_ind);
double * CGS_randomB_traceInvAA0Fx(int taille, int nombre, double *val,int *row_ptr,int *col_ind,const CImg<double> &IxIx,const CImg<double> &IxIy,const CImg<double> &IyIy,const CImg<double> &data_out,const CImgList<double> &reg_out,const CImg<double> &x, int nbIt, double tol, bool periodic);
double CGS_randomB_traceInvAA0(int taille, int nombre, double *val,int *row_ptr,int *col_ind,const CImg<double> &IxIx,const CImg<double> &IxIy,const CImg<double> &IyIy,const CImg<double> &data_out, int nbIt, double tol);
double CGS_randomB_traceInvA(int taille, int nombre,int nc, int nl, double *val,int *row_ptr,int *col_ind, int nbIt, double tol);	
	/*************************************************************************************************************/

	/*************************************************************************************************************/
	/*********************************************Structure Functions*********************************************/
	/*************************************************************************************************************/

	/*************************************************************************************************************/
CImg<double> StructureFunction2nd_horiVertDir(const CImg<double> &UV, int neta, int lo,const CImgList <double> &z_right, const CImgList <double> &z_down,const CImgList <double> &z_t_right, const CImgList <double> &z_t_down);
CImg<double> StructureFunction2nd_horiVertDir(const CImg<double> &UV, int neta, int lo, bool trans);
CImg<double> StructureFunction2ndTrans_horiVertDir(const CImg<double> &UV, int neta, int lo);
CImg<double> StructureFunction2ndLong_horiVertDir(const CImg<double> &UV, int neta, int lo);
CImg<double> StructureFunctionPth_horiVertDir(const CImg<double> &UV, int neta, int lo,const CImgList <double> &z_right, const CImgList <double> &z_down,const CImgList <double> &z_t_right, const CImgList <double> &z_t_down,double p);
CImg<double> StructureFunctionPth_horiVertDir(const CImg<double> &UV, int neta, int lo, bool trans,double p);
CImg<double> StructureFunctionPthTrans_horiVertDir(const CImg<double> &UV, int neta, int lo,double p);
CImg<double> StructureFunctionPthLong_horiVertDir(const CImg<double> &UV, int neta, int lo,double p);
CImg<double> StructureFunctionPthAbs_horiVertDir(const CImg<double> &UV, int neta, int lo,const CImgList <double> &z_right, const CImgList <double> &z_down,const CImgList <double> &z_t_right, const CImgList <double> &z_t_down,double p);
CImg<double> StructureFunctionPthAbs_horiVertDir(const CImg<double> &UV, int neta, int lo, bool trans,double p);
CImg<double> StructureFunctionPthTransAbs_horiVertDir(const CImg<double> &UV, int neta, int lo,double p);
CImg<double> StructureFunctionPthLongAbs_horiVertDir(const CImg<double> &UV, int neta, int lo,double p);
CImg<double> StructureFunction2nd_crossDir(const CImg<double> &UV, int neta, int lo,const CImgList <double> &z_rightDown, const CImgList <double> &z_rightUp,const CImgList <double> &z_t_rightDown, const CImgList <double> &z_t_rightUp);
CImg<double> StructureFunction2nd_crossDir(const CImg<double> &UV, int neta, int lo, bool trans);
CImg<double> StructureFunction2ndTrans_crossDir(const CImg<double> &UV, int neta, int lo);
CImg<double> StructureFunction2ndLong_crossDir(const CImg<double> &UV, int neta, int lo);
CImg<double> StructureFunctionPth_crossDir(const CImg<double> &UV, int neta, int lo,const CImgList <double> &z_rightDown, const CImgList <double> &z_rightUp,const CImgList <double> &z_t_rightDown, const CImgList <double> &z_t_rightUp, double p);
CImg<double> StructureFunctionPth_crossDir(const CImg<double> &UV, int neta, int lo, bool trans,double p);
CImg<double> StructureFunctionPthTrans_crossDir(const CImg<double> &UV, int neta, int lo, double p);
CImg<double> StructureFunctionPthLong_crossDir(const CImg<double> &UV, int neta, int lo, double p);
CImg<double> StructureFunctionPthAbs_crossDir(const CImg<double> &UV, int neta, int lo,const CImgList <double> &z_rightDown, const CImgList <double> &z_rightUp,const CImgList <double> &z_t_rightDown, const CImgList <double> &z_t_rightUp, double p);
CImg<double> StructureFunctionPthAbs_crossDir(const CImg<double> &UV, int neta, int lo, bool trans,double p);
CImg<double> StructureFunctionPthTransAbs_crossDir(const CImg<double> &UV, int neta, int lo, double p);
CImg<double> StructureFunctionPthLongAbs_crossDir(const CImg<double> &UV, int neta, int lo, double p);
	/*************************************************************************************************************/

	/*************************************************************************************************************/
	/*********************************************Sub gradients***************************************************/
	/*************************************************************************************************************/

	/*************************************************************************************************************/
void subGradients_K41_horiVertDir(CImg<double> &subGrad, const CImg<double> &UV, double *betaFunc_HV, int LMin, int LMax,const CImgList <double> &z_right, const CImgList <double> &z_down,const CImgList <double> &z_t_right, const CImgList <double> &z_t_down);
void subGradients_K41_crossDir(CImg<double> &subGrad, const CImg<double> &UV, double *betaFunc_cross, int LMin, int LMax, const CImgList <double> &z_rightDown, const CImgList <double> &z_rightUp, const CImgList <double> &z_t_rightDown, const CImgList <double> &z_t_rightUp);
void subGradients_K41_horiVertDir(CImg<double> &subGrad, const CImg<double> &UV, double beta, double power, int LMin, int LMax, bool trans);
void subGradients_K41_crossDir(CImg<double> &subGrad, const CImg<double> &UV, double beta, double power, int LMin, int LMax, bool trans);  	
double ConstraintsAbsDev(const CImg<double> &UV,  double beta,double power, int neta, int lo);
double ConstraintsAbsDevLongTrans(const CImg<double> &UV,  double beta,double power, int neta, int lo);
double ConstraintsAbsDevLong(const CImg<double> &UV,  double beta,double power, int neta, int lo);
	/*************************************************************************************************************/

	/*************************************************************************************************************/
	/*********************************************Horn and Schunck      ******************************************/
	/*************************************************************************************************************/

	/*************************************************************************************************************/
bool HornAndSchunck(CImg<double> &UV,const CImg<double> &UVtil,const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> &im_t,float alpha,const CImg<double> &data_out, bool periodic);
bool HornAndSchunck(CImg<double> &UV,const CImg<double> &UVtil,const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> &im_t,float alpha,const CImg<double> &data_out,const CImgList<double> &reg_out, bool periodic);
//bool HornAndSchunckOld(CImg<double> &UV,const CImg<double> &UVtil,const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> &im_t,float alpha,const CImg<double> data_out, bool periodic);
	/*************************************************************************************************************/

	/*************************************************************************************************************/
	/***************************************K41   Quadratic form filling******************************************/
	/*************************************************************************************************************/

	/*************************************************************************************************************/
double Constant_C(const CImg<double> &ItIt,const CImg<double> &data_out,const CImg<double> &Util,const CImg<double> &Vtil,const CImg<double> &lambda_S2,const CImg<double> &lambda_S2_cross,double* betaFunc_HV, double *betaFunc_cross, const CImgList <double> &z_right, const CImgList <double> &z_down, const CImgList <double> &z_rightDown, const CImgList <double> &z_rightUp, const CImgList <double> &z_t_right, const CImgList <double> &z_t_down, const CImgList <double> &z_t_rightDown, const CImgList <double> &z_t_rightUp,int neta, int lo,int nc,int nl);
void  fillingVector_B(VECTOR_double &B,const CImg<double> &ItIx,const CImg<double> &ItIy,const CImg<double> &Util,const CImg<double> &Vtil,const CImg<double> &data_out,const CImg<double> &lambda_S2,const CImg<double> &lambda_S2_cross,const CImgList <double> &z_right,const CImgList <double> &z_down,const CImgList <double> &z_rightDown,const CImgList <double> &z_rightUp,const CImgList <double> &z_t_right,const CImgList <double> &z_t_down,const CImgList <double> &z_t_rightDown,const CImgList <double> &z_t_rightUp,int neta,int lo,int nc,int nl);
void fillingSparseHessian_A(double *val,int *col_ind,int *row_ptr,const CImg<double> &IxIx,const CImg<double> &IxIy,const CImg<double> &IyIy,const CImg<double> &data_out,const CImg<double> &lambda_S2,const CImg<double> &lambda_S2_cross,const CImgList <double> &z_right,const CImgList <double> &z_down,const CImgList <double> &z_rightDown,const CImgList <double> &z_rightUp,const CImgList <double> &z_t_right,const CImgList <double> &z_t_down,const CImgList <double> &z_t_rightDown,const CImgList <double> &z_t_rightUp,int neta,int lo,int nc,int nl,int nombre);
	/*************************************************************************************************************/

	/*************************************************************************************************************/
	/*********************************************Dual Function***************************************************/
	/*************************************************************************************************************/

	/*************************************************************************************************************/
double DualFunction(const CImg<double> &UV, const CImg<double> &UVtil,const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> &im_t,const CImg<double> data_out, const CImg<double> &lambda_S2, const CImg<double> &lambda_S2_cross,const CImgList <double> &z_right, const CImgList <double> &z_down, const CImgList <double> &z_rightDown, const CImgList <double> &z_rightUp,const CImgList <double> &z_t_right, const CImgList <double> &z_t_down, const CImgList <double> &z_t_rightDown, const CImgList <double> &z_t_rightUp,float scaleSmooth, int neta, int lo, double* betaFunc_HV, double *betaFunc_cross,int nombre);
double DualFunction_Gen_Rob_BruhnAndAl(const CImg<double> &UV, const CImg<double> &UVtil,const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> &im_t,const CImg<double> data_out, const CImg<double> &lambda_S2, const CImg<double> &lambda_S2_cross,const CImgList <double> &z_right, const CImgList <double> &z_down, const CImgList <double> &z_rightDown, const CImgList <double> &z_rightUp,const CImgList <double> &z_t_right, const CImgList <double> &z_t_down, const CImgList <double> &z_t_rightDown, const CImgList <double> &z_t_rightUp,float scaleSmooth, int lo, double* betaFunc_HV, double *betaFunc_cross,int nombre);
double DualFunction_Gen_BruhnAndAl(const CImg<double> &UV, const CImg<double> &UVtil,const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> &im_t,const CImg<double> data_out, const CImg<double> &lambda_S2, const CImg<double> &lambda_S2_cross,float scaleSmooth, int lo, double* betaFunc_HV, double *betaFunc_cross,int nombre);
double DualFunction_K41LongTrans(const CImg<double> &UV, const CImg<double> &UVtil,const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> &im_t,const CImg<double> data_out, const CImg<double> &lambda_S2, const CImg<double> &lambda_S2_cross,int neta, int lo, double beta, double power,int nombre);
double DualFunction_K41Long(const CImg<double> &UV, const CImg<double> &UVtil,const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> &im_t,const CImg<double> data_out, const CImg<double> &lambda_S2, const CImg<double> &lambda_S2_cross,int neta, int lo, double beta, double power,int nombre);
double DualFunction_K41Long(const CImg<double> &UV, const CImg<double> &UVtil,const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> &im_t,const CImg<double> data_out, const CImg<double> &lambda_S2, const CImg<double> &lambda_S2_cross,float scaleSmooth,int neta, int lo, double beta, double power,int nombre);
double DataEnergy(const CImg<double> &U, const CImg<double> &V,const CImg<double> &IxIx,const CImg<double> &IyIy,const CImg<double> &IxIy,const CImg<double> &ItIx,const CImg<double> &ItIy,const CImg<double> &ItIt,const CImg<double> data_out);
	/*************************************************************************************************************/

	/*************************************************************************************************************/
	/*******************************************Primal Function***************************************************/
	/*************************************************************************************************************/

	/*************************************************************************************************************/
bool PrimalMinimizer(CImg<double> &UV,const CImg<double> &UVtil,const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> &im_t,const CImg<double> &data_out, const CImg<double> &lambda_S2,const CImg<double> &lambda_S2_cross,const CImgList <double> &z_right, const CImgList <double> &z_down, const CImgList <double> &z_rightDown, const CImgList <double> &z_rightUp,const CImgList <double> &z_t_right, const CImgList <double> &z_t_down, const CImgList <double> &z_t_rightDown, const CImgList <double> &z_t_rightUp,float scaleSmooth,int neta,  int lo, int nombre);
bool PrimalMinimizer_Gen_Rob_BruhnAndAl(CImg<double> &UV, const CImg<double> &UVtil,const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> &im_t,const CImg<double> data_out, const CImg<double> &lambda_S2, const CImg<double> &lambda_S2_cross,const CImgList <double> &z_right, const CImgList <double> &z_down, const CImgList <double> &z_rightDown, const CImgList <double> &z_rightUp,const CImgList <double> &z_t_right, const CImgList <double> &z_t_down, const CImgList <double> &z_t_rightDown, const CImgList <double> &z_t_rightUp,float scaleSmooth, int lo, int nombre);
bool PrimalMinimizer_Gen_BruhnAndAl(CImg<double> &UV, const CImg<double> &UVtil,const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> &im_t,const CImg<double> data_out, const CImg<double> &lambda_S2, const CImg<double> &lambda_S2_cross,float scaleSmooth, int lo,int nombre);
bool PrimalMinimizer_K41LongTrans(CImg<double> &UV, const CImg<double> &UVtil,const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> &im_t,const CImg<double> data_out, const CImg<double> &lambda_S2, const CImg<double> &lambda_S2_cross,int neta, int lo,int nombre);
bool PrimalMinimizer_K41Long(CImg<double> &UV, const CImg<double> &UVtil,const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> &im_t,const CImg<double> data_out, const CImg<double> &lambda_S2, const CImg<double> &lambda_S2_cross,int neta, int lo,int nombre);
bool PrimalMinimizer_K41Long(CImg<double> &UV, const CImg<double> &UVtil,const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> &im_t,const CImg<double> data_out, const CImg<double> &lambda_S2, const CImg<double> &lambda_S2_cross,float scaleSmooth,int neta, int lo,int nombre);
	/*************************************************************************************************************/

	/*************************************************************************************************************/
	/********************************Number element in sparse Hessian Matrix**************************************/
	/*************************************************************************************************************/

	/*************************************************************************************************************/
int NumberNonZeroElementInHessian(const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> &data_out, const CImg<double> &lambda_S2,const CImg<double> &lambda_S2_cross, const CImgList <double> &z_right, const CImgList <double> &z_down, const CImgList <double> &z_rightDown, const CImgList <double> &z_rightUp, const CImgList <double> &z_t_right, const CImgList <double> &z_t_down, const CImgList <double> &z_t_rightDown, const CImgList <double> &z_t_rightUp,float scaleSmooth,int neta, int lo);
int NumberNonZeroElementInHessian_Gen_Rob_BruhnAndAl(const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> data_out, const CImg<double> &lambda_S2, const CImg<double> &lambda_S2_cross,const CImgList <double> &z_right, const CImgList <double> &z_down, const CImgList <double> &z_rightDown, const CImgList <double> &z_rightUp,const CImgList <double> &z_t_right, const CImgList <double> &z_t_down, const CImgList <double> &z_t_rightDown, const CImgList <double> &z_t_rightUp,float scaleSmooth, int lo);
int NumberNonZeroElementInHessian_Gen_BruhnAndAl(const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> data_out, const CImg<double> &lambda_S2, const CImg<double> &lambda_S2_cross,float scaleSmooth, int lo);
int NumberNonZeroElementInHessian_K41LongTrans(const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> data_out, const CImg<double> &lambda_S2, const CImg<double> &lambda_S2_cross,int neta, int lo);
int NumberNonZeroElementInHessian_K41Long(const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> data_out, const CImg<double> &lambda_S2, const CImg<double> &lambda_S2_cross,int neta, int lo);
int NumberNonZeroElementInHessian_K41Long(const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> data_out, const CImg<double> &lambda_S2, const CImg<double> &lambda_S2_cross,float scaleSmooth,int neta, int lo);
	/*************************************************************************************************************/

	/*************************************************************************************************************/
	/********************************Log determinant of sparse Hessian Matrix*************************************/
	/*************************************************************************************************************/

	/*************************************************************************************************************/
double logdetA(const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> &im_t,const CImg<double> data_out, const CImg<double> &lambda_S2,const CImg<double> &lambda_S2_cross,const CImgList <double> &z_right, const CImgList <double> &z_down, const CImgList <double> &z_rightDown, const CImgList <double> &z_rightUp,const CImgList <double> &z_t_right, const CImgList <double> &z_t_down, const CImgList <double> &z_t_rightDown, const CImgList <double> &z_t_rightUp,float scaleSmooth, int neta, int lo, int nombre);
double logdetA_Gen_Rob_BruhnAndAl(const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> &im_t,const CImg<double> data_out, const CImg<double> &lambda_S2, const CImg<double> &lambda_S2_cross,const CImgList <double> &z_right, const CImgList <double> &z_down, const CImgList <double> &z_rightDown, const CImgList <double> &z_rightUp,const CImgList <double> &z_t_right, const CImgList <double> &z_t_down, const CImgList <double> &z_t_rightDown, const CImgList <double> &z_t_rightUp,float scaleSmooth, int lo, int nombre);
double logdetA_Gen_BruhnAndAl(const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> &im_t,const CImg<double> data_out, const CImg<double> &lambda_S2, const CImg<double> &lambda_S2_cross,float scaleSmooth, int lo,int nombre);
double logdetA_K41LongTrans(const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> &im_t,const CImg<double> data_out, const CImg<double> &lambda_S2, const CImg<double> &lambda_S2_cross,int neta, int lo,int nombre);
double logdetA_K41Long(const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> &im_t,const CImg<double> data_out, const CImg<double> &lambda_S2, const CImg<double> &lambda_S2_cross,int neta, int lo,int nombre);
	/*************************************************************************************************************/

	/*************************************************************************************************************/
	/********************************Evidence of beta and power law    *******************************************/
	/*************************************************************************************************************/

	/*************************************************************************************************************/
double evidence_beta_Gen_Rob_BruhnAndAl(const CImg<double> &UV,const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> &im_t,const CImg<double> &data_out,const CImg<double> &lambda_S2,const CImg<double> &lambda_S2_cross,const CImgList <double> &z_right,const CImgList <double> &z_down,const CImgList <double> &z_rightDown,const CImgList <double> &z_rightUp,const CImgList <double> &z_t_right,const CImgList <double> &z_t_down,const CImgList <double> &z_t_rightDown,const CImgList <double> &z_t_rightUp,float scaleSmooth,int neta,int lo, int nombre,int nbIt, double tol, double s);
double evidence_beta_Gen_BruhnAndAl(const CImg<double> &UV,const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> &im_t,const CImg<double> &data_out,const CImg<double> &lambda_S2,const CImg<double> &lambda_S2_cross,float scaleSmooth, int neta,int lo, int nombre,int nbIt, double tol, double s);
double evidence_betaLong(const CImg<double> &UV,const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> &im_t,const CImg<double> &data_out,const CImg<double> &lambda_S2,const CImg<double> &lambda_S2_cross,int neta,int lo, int nombre,int nbIt, double tol, double s);
double evidence_betaLong(const CImg<double> &UV,const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> &im_t,const CImg<double> &data_out,const CImg<double> &lambda_S2,const CImg<double> &lambda_S2_cross,float scaleSmooth,int neta,int lo, int nombre,int nbIt, double tol, double s);
double evidence_betaLongTrans(const CImg<double> &UV,const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> &im_t,const CImg<double> &data_out,const CImg<double> &lambda_S2,const CImg<double> &lambda_S2_cross,int neta,int lo, int nombre,int nbIt, double tol, double s);
double evidence_powerLaw_Gen_Rob_BruhnAndAl(const CImg<double> &UV,const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> &im_t,const CImg<double> &data_out,double beta,const CImg<double> &lambda_S2,const CImg<double> &lambda_S2_cross,const CImgList <double> &z_right,const CImgList <double> &z_down,const CImgList <double> &z_rightDown,const CImgList <double> &z_rightUp,const CImgList <double> &z_t_right,const CImgList <double> &z_t_down,const CImgList <double> &z_t_rightDown,const CImgList <double> &z_t_rightUp,float scaleSmooth,int neta,int lo, int nombre, double s, CImg<double> &save);
double evidence_powerLawLong(const CImg<double> &UV,const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> &im_t,const CImg<double> &data_out,double beta,const CImg<double> &lambda_S2,const CImg<double> &lambda_S2_cross,int neta,int lo, int nombre, double s);
double evidence_powerLawLong(const CImg<double> &UV,const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> &im_t,const CImg<double> &data_out,double beta,const CImg<double> &lambda_S2,const CImg<double> &lambda_S2_cross,float scaleSmooth, int neta,int lo, int nombre, double s);
double evidence_powerLawLongTrans(const CImg<double> &UV,const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> &im_t,const CImg<double> &data_out,double beta,const CImg<double> &lambda_S2,const CImg<double> &lambda_S2_cross,int neta,int lo, int nombre, double s);
double evidence_powerLawLong(const CImg<double> &UV,const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> &im_t,const CImg<double> &data_out,double beta,const CImg<double> &lambda_S2,const CImg<double> &lambda_S2_cross,int neta,int lo, int nombre, double s, CImg<double> &save);
double evidence_powerLawLong(const CImg<double> &UV,const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> &im_t,const CImg<double> &data_out,double beta,const CImg<double> &lambda_S2,const CImg<double> &lambda_S2_cross,int scaleSmooth,int neta,int lo, int nombre, double s, CImg<double> &save);
double evidence_powerLawLongTrans(const CImg<double> &UV,const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> &im_t,const CImg<double> &data_out,double beta,const CImg<double> &lambda_S2,const CImg<double> &lambda_S2_cross,int neta,int lo, int nombre, double s, CImg<double> &save);
/*************************************************************************************************************************************/

/*************************************************************************************************************************************/
/**********************Optimization Dual function (Quasi nexton + Wolf conditions)****************************************************/
/*************************************************************************************************************************************/

/*************************************************************************************************************************************/
double optimizeDualFunction_LongK41(CImg<double> &lambda, CImg<double> &lambda_cross,  CImg<double> &Xinc,const CImg<double> &Xtil,const CImg<double> &trueUV,const CImg<double> &f_x,const CImg<double> &f_y,const CImg<double> &f_t, CImg<double> &data_out,const CImg<double> &mask, int resol,int nombre,int offEdge,int dxOrig, int dyOrig, int norm, double sigma_obs, double gamma, double zeta,float scaleSmooth,int neta,  int lo, double esp,int nbIt_lbfgs,  bool displayMode, bool DivZero);



#endif
